"""Implementation of ClientForm interface in terms of HTML DOM level 2.

For full copyright information and license terms, see the file COPYING included
with the distribution.

The code in this file free software; you can redistribute it and/or modify
it under the terms of the BSD License.  Copyright 2003 John J. Lee
<jjl@pobox.com>

Comments to jjl@pobox.com

"""

# XXX
# Don't .remove items from *_attrs -- only .append.

# Differences from htmllib implementation:
#  action attribute isn't an absolute URI
#  TEXTAREA control values include leading and trailing whitespace
#   (this is probably a bug in ClientForm 0.0.x & 0.1.x)
#  most strings are unicode (should they all be? -- eg. u"submitbutton"?)
#  ListControl.multiple is writable -- there are probably a few other
#   differences like this one

VERSION = "0.0.1a-pre2"

import copy, urllib, urllib2, mimetools
from cStringIO import StringIO

from urlparse import urljoin
from types import TupleType

def isstringlike(x):
    try: x+""
    except: return False
    else: return True

# This cut-n-pasted MimeWriter from standard library is here so can add
# to HTTP headers rather than message body when appropriate.  It also uses
# \r\n in place of \n.  This is nasty.
class MimeWriter:

    """Generic MIME writer.

    Methods:

    __init__()
    addheader()
    flushheaders()
    startbody()
    startmultipartbody()
    nextpart()
    lastpart()

    A MIME writer is much more primitive than a MIME parser.  It
    doesn't seek around on the output file, and it doesn't use large
    amounts of buffer space, so you have to write the parts in the
    order they should occur on the output file.  It does buffer the
    headers you add, allowing you to rearrange their order.

    General usage is:

    f = <open the output file>
    w = MimeWriter(f)
    ...call w.addheader(key, value) 0 or more times...

    followed by either:

    f = w.startbody(content_type)
    ...call f.write(data) for body data...

    or:

    w.startmultipartbody(subtype)
    for each part:
        subwriter = w.nextpart()
        ...use the subwriter's methods to create the subpart...
    w.lastpart()

    The subwriter is another MimeWriter instance, and should be
    treated in the same way as the toplevel MimeWriter.  This way,
    writing recursive body parts is easy.

    Warning: don't forget to call lastpart()!

    XXX There should be more state so calls made in the wrong order
    are detected.

    Some special cases:

    - startbody() just returns the file passed to the constructor;
      but don't use this knowledge, as it may be changed.

    - startmultipartbody() actually returns a file as well;
      this can be used to write the initial 'if you can read this your
      mailer is not MIME-aware' message.

    - If you call flushheaders(), the headers accumulated so far are
      written out (and forgotten); this is useful if you don't need a
      body part at all, e.g. for a subpart of type message/rfc822
      that's (mis)used to store some header-like information.

    - Passing a keyword argument 'prefix=<flag>' to addheader(),
      start*body() affects where the header is inserted; 0 means
      append at the end, 1 means insert at the start; default is
      append for addheader(), but insert for start*body(), which use
      it to determine where the Content-type header goes.

    """

    def __init__(self, fp, http_hdrs=None):
        self._http_hdrs = http_hdrs
        self._fp = fp
        self._headers = []
        self._boundary = []
        self._first_part = True

    def addheader(self, key, value, prefix=0,
                  add_to_http_hdrs=0):
        """
        prefix is ignored if add_to_http_hdrs is true.
        """
        lines = value.split("\r\n")
        while lines and not lines[-1]: del lines[-1]
        while lines and not lines[0]: del lines[0]
        if add_to_http_hdrs:
            value = "".join(lines)
            self._http_hdrs.append((key, value))
        else:
            for i in range(1, len(lines)):
                lines[i] = "    " + lines[i].strip()
            value = "\r\n".join(lines) + "\r\n"
            line = key + ": " + value
            if prefix:
                self._headers.insert(0, line)
            else:
                self._headers.append(line)

    def flushheaders(self):
        self._fp.writelines(self._headers)
        self._headers = []

    def startbody(self, ctype=None, plist=[], prefix=1,
                  add_to_http_hdrs=0, content_type=1):
        """
        prefix is ignored if add_to_http_hdrs is true.
        """
        if content_type and ctype:
            for name, value in plist:
                ctype = ctype + ';\r\n %s=\"%s\"' % (name, value)
            self.addheader("Content-type", ctype, prefix=prefix,
                           add_to_http_hdrs=add_to_http_hdrs)
        self.flushheaders()
        if not add_to_http_hdrs: self._fp.write("\r\n")
        self._first_part = True
        return self._fp

    def startmultipartbody(self, subtype, boundary=None, plist=[], prefix=1,
                           add_to_http_hdrs=0, content_type=1):
        boundary = boundary or mimetools.choose_boundary()
        self._boundary.append(boundary)
        return self.startbody("multipart/" + subtype,
                              [("boundary", boundary)] + plist,
                              prefix=prefix,
                              add_to_http_hdrs=add_to_http_hdrs,
                              content_type=content_type)

    def nextpart(self):
        boundary = self._boundary[-1]
        if self._first_part:
            self._first_part = False
        else:
            self._fp.write("\r\n")
        self._fp.write("--" + boundary + "\r\n")
        return self.__class__(self._fp)

    def lastpart(self):
        if self._first_part:
            self.nextpart()
        boundary = self._boundary.pop()
        self._fp.write("\r\n--" + boundary + "--\r\n")


class ControlNotFoundError(ValueError): pass
class ItemNotFoundError(ValueError): pass
class ItemCountError(ValueError): pass


def attrs_dict_from_node(node):
    attrs = {}
    for key, value in node.attributes.iteritems():
        if isinstance(key, TupleType):
            key = key[1]  # discard namespace
        key = key.lower()
        attrs[key] = value.nodeValue
    return attrs

def button_type(node):
    ctype = node.type.lower()
    if ctype in ["", "submit"]: ctype = "submitbutton"
    elif ctype == "reset": ctype = "resetbutton"
    elif ctype == "button": ctype = "buttonbutton"
    else: assert False, "unknown BUTTON type '%s'" % cn.type
    return ctype


class ReadonlyMixin:
    readonly_attrs = []
    method_attrs = []
    def __getattr__(self, name):
        if name in self.method_attrs:
            meth_name = "_get_"+name
            return getattr(self, meth_name)()
        return getattr(self.__class__, name)
    def __setattr__(self, name, value):
        if name in self.__class__.readonly_attrs:
            raise AttributeError("attribute %s is read-only" % name)
        if name in self.method_attrs:
            meth_name = "_set_"+name
            getattr(self, meth_name)(value)
        else:
            self.__dict__[name] = value


class Control:
    """An HTML form control.

    An HTMLForm contains a sequence of Controls.  HTMLForm delegates lots of
    things to Control objects, and most of Control's methods are, in effect,
    documented by the HTMLForm docstrings.

    The Controls in an HTMLForm can be got at via the HTMLForm.find_control
    method or the HTMLForm.controls attribute.

    Control instances are usually constructed using the ParseFile /
    ParseResponse functions, so you can probably ignore the rest of this
    paragraph.  A Control is only properly initialised after the fixup method
    has been called.  In fact, this is only strictly necessary for ListControl
    instances.  This is necessary because ListControls are built up from
    ListControls each containing only a single item, and their initial value(s)
    can only be known after the sequence is complete.

    The types and values that are acceptable for assignment to the value
    attribute are defined by subclasses.

    If the disabled attribute is true, this represents the state typically
    represented by browsers by `greying out' a control.  If the disabled
    attribute is true, the Control will raise AttributeError if an attempt is
    made to change its value.  In addition, the control will not be considered
    `successful' as defined by the W3C HTML 4 standard -- ie. it will
    contribute no data to the return value of the HTMLForm.click* methods.  To
    enable a control, set the disabled attribute to a false value.

    If the readonly attribute is true, the Control will raise AttributeError if
    an attempt is made to change its value.  To make a control writable, set
    the readonly attribute to a false value.

    All controls have the disabled anre readonly attribute, not only those that
    may have the HTML attributes of the same names.

    On assignment to the value attribute, the following exceptions are raised:
    TypeError, AttributeError (if the value attribute should not be assigned
    to, because the control is disabled, for example) and ValueError.

    If the name or value attributes are None, or the value is an empty list, or
    if the control is disabled, the control is not successful.

    Public attributes:

    type: string describing type of control (see the keys of the
     HTMLForm.type2class dictionary for the allowable values) (readonly)
    name: name of control (readonly)
    value: current value of control (subclasses may allow a single value, a
     sequence of values, or either)
    disabled: disabled state
    readonly: readonly state
    id: value of id HTML attribute

    """
    def __init__(self, type, name, attrs):
        """
        type: string describing type of control (see the keys of the
         HTMLForm.type2class dictionary for the allowable values)
        name: control name
        attrs: HTML attributes of control's HTML element

        """
        raise NotImplementedError()

    def add_to_form(self, form):
        form.controls.append(self)

    def fixup(self):
        pass

    def __getattr__(self, name): raise NotImplementedError()
    def __setattr__(self, name, value): raise NotImplementedError()

    def pairs(self):
        """Return list of (key, value) pairs suitable for passing to urlencode.
        """
        raise NotImplementedError()

    def _write_mime_data(self, mw):
        """Write data for this control to a MimeWriter."""
        for name, value in self.pairs():
            mw2 = mw.nextpart()
            mw2.addheader("Content-disposition",
                          'form-data; name="%s"' % name, 1)
            f = mw2.startbody()
            f.write(value)

    def __str__(self):
        raise NotImplementedError()


class ScalarControl(ReadonlyMixin, Control):
    """Control whose value is not restricted to one of a prescribed set.

    Some ScalarControls don't accept any value attribute.  Otherwise, takes a
    single value, which must be string-like.

    Additional read-only public attribute:

    attrs: dictionary mapping the names of original HTML attributes of the
     control to their values

    """
    readonly_attrs = ["type", "name", "id", "attrs"]
    method_attrs = ["type", "name", "id", "value", "disabled", "attrs"]
    def __init__(self, node):
        self.node = node
        self._clicked = False
    def _get_type(self):
        return self.node.getAttribute("type").lower()
    def _get_name(self):
        if self.node.hasAttribute("name"):
            return self.node.getAttribute("name")
        else:
            return None
    def _get_id(self):
        if self.node.hasAttribute("id"):
            return self.node.getAttribute("id")
        else:
            return None
    def _get_value(self):
        if self.node.hasAttribute("value"):
            return self.node.getAttribute("value")
        else:
            return None
    def _set_value(self, value):
        if not isstringlike(value):
            raise TypeError("must assign a string")
        elif self.readonly:
            raise AttributeError("control '%s' is readonly" % self.name)
        elif self.disabled:
            raise AttributeError("control '%s' is disabled" % self.name)
        if value is None:
            self.node.removeAttribute("value")
        else:
            self.node.value = value
    def _get_disabled(self):
        return self.node.disabled
    def _set_disabled(self, value):
        self.node.disabled = value
    def _get_attrs(self):
        return attrs_dict_from_node(self.node)

    def pairs(self):
        name = self.name
        value = self.value
        if name is None or value is None or self.disabled:
            return []
        return [(name, value)]

    def __str__(self):
        name = self.name
        value = self.value
        if name is None: name = "<None>"
        if value is None: value = "<None>"

        infos = []
        if self.disabled: infos.append("disabled")
        if self.readonly: infos.append("readonly")
        info = ", ".join(infos)
        if info: info = " (%s)" % info

        return "<%s(%s=%s)%s>" % (self.__class__.__name__, name, value, info)


class FileControl(ScalarControl):
    """File upload with INPUT TYPE=FILE.

    The value attribute of a FileControl is always None.  Use add_file instead.

    Additional public method: add_file

    """
    readonly_attrs = copy.copy(ScalarControl.readonly_attrs)
    readonly_attrs.append("value")
    def __init__(self, node):
        ScalarControl.__init__(self, node)
        self._upload_data = []
        self.readonly = False

    def _get_value(self): return None

    def add_file(self, file_object, content_type=None, filename=None):
        if not hasattr(file_object, "read"):
            raise TypeError("file-like object must have read method")
        if content_type is not None and not isstringlike(content_type):
            raise TypeError("content type must be None or string-like")
        if filename is not None and not isstringlike(filename):
            raise TypeError("filename must be None or string-like")
        if content_type is None:
            content_type = "application/octet-stream"
        self._upload_data.append((file_object, content_type, filename))

    def pairs(self):
        # XXX should it be successful even if unnamed?
        if self.name is None or self.disabled:
            return []
        return [(self.name, "")]

    def _write_mime_data(self, mw):
        # called by HTMLForm
        if len(self._upload_data) == 1:
            # single file
            file_object, content_type, filename = self._upload_data[0]
            mw2 = mw.nextpart()
            fn_part = filename and ('; filename="%s"' % filename) or ''
            disp = 'form-data; name="%s"%s' % (self.name, fn_part)
            mw2.addheader("Content-disposition", disp, prefix=1)
            fh = mw2.startbody(content_type, prefix=0)
            fh.write(file_object.read())
        elif len(self._upload_data) != 0:
            # multiple files
            mw2 = mw.nextpart()
            disp = 'form-data; name="%s"' % self.name
            mw2.addheader("Content-disposition", disp, prefix=1)
            fh = mw2.startmultipartbody("mixed", prefix=0)
            for file_object, content_type, filename in self._upload_data:
                mw3 = mw2.nextpart()
                fn_part = filename and ('; filename="%s"' % filename) or ''
                disp = 'file%s' % fn_part
                mw3.addheader("Content-disposition", disp, prefix=1)
                fh2 = mw3.startbody(content_type, prefix=0)
                fh2.write(file_object.read())
            mw2.lastpart()

    def __str__(self):
        name = self.name
        if name is None: name = "<None>"

        if not self._upload_data:
            value = "<No files added>"
        else:
            value = []
            for file, ctype, filename in self._upload_data:
                if filename is None:
                    value.append("<Unnamed file>")
                else:
                    value.append(filename)
            value = ", ".join(value)

        info = []
        if self.disabled: info.append("disabled")
        if self.readonly: info.append("readonly")
        info = ", ".join(info)
        if info: info = " (%s)" % info

        return "<%s(%s=%s)%s>" % (self.__class__.__name__, name, value, info)


class HiddenControl(ScalarControl):
    def __init__(self, node):
        ScalarControl.__init__(self, node)
        self.readonly = False
        if self.value is None: self.value = ""
        self.readonly = True


class TextControl(ScalarControl):
    """Textual input control.

    Covers:

    INPUT/TEXT
    INPUT/PASSWORD
    INPUT/FILE
    INPUT/HIDDEN
    TEXTAREA

    """
    method_attrs = copy.copy(ScalarControl.method_attrs)
    method_attrs.append("readonly")
    def __init__(self, node):
        ScalarControl.__init__(self, node)
        if self.value is None: self.value = ""
        self.readonly = False
    def _get_readonly(self):
        return self.node.readOnly
    def _set_readonly(self, value):
        self.node.readOnly = value


class PasswordControl(TextControl): pass


class TextareaControl(TextControl):
    def __init__(self, node):
        TextControl.__init__(self, node)
    def _get_type(self): return "textarea"
    def _get_readonly(self):
        return self.node.readonly
    def _set_readonly(self, value):
        self.node.readonly = value
    def _get_value(self):
        return self.node.value
    def _set_value(self, value):
        if not isstringlike(value):
            raise TypeError("must assign a string")
        elif self.readonly:
            raise AttributeError("control '%s' is readonly" % self.name)
        elif self.disabled:
            raise AttributeError("control '%s' is disabled" % self.name)
        if value is None:
            for node in self.node.childNodes:
                self.node.removeChild(node)
        else:
            self.node.value = value


class IsindexControl(ScalarControl):
    """ISINDEX control.

    ISINDEX is the odd-one-out of HTML form controls.  In fact, it isn't really
    part of regular HTML forms at all, and predates it.  You're only allowed
    one ISINDEX per HTML document.  ISINDEX and regular form submission are
    mutually exclusive -- either submit a form, or the ISINDEX.

    Having said this, since ISINDEX controls may appear in forms (which is
    probably bad HTML), ParseFile / ParseResponse will include them in the
    HTMLForm instances it returns.  You can set the ISINDEX's value, as with
    any other control (but note that ISINDEX controls have no name, so you'll
    need to use the type argument of set_value!).  When you submit the form,
    the ISINDEX will not be successful (ie., no data will get returned to the
    server as a result of its presence), unless you click on the ISINDEX
    control, in which case the ISINDEX gets submitted instead of the form:

    form.set_value("my isindex value", type="isindex")
    urllib2.urlopen(form.click(type="isindex"))

    ISINDEX elements outside of FORMs are ignored.  If you want to submit one
    by hand, do it like so:

    url = urlparse.urljoin(page_uri, "?"+urllib.quote_plus("my isindex value"))
    result = urllib2.urlopen(url)

    """
    method_attrs = copy.copy(ScalarControl.method_attrs)
    method_attrs.remove("disabled")
    def __init__(self, node):
        ScalarControl.__init__(self, node)
        self._value = ""
        self.disabled = False
        self.readonly = False
    def _get_type(self): return "isindex"
    def _get_value(self): return self._value
    def _set_value(self, value):
        if not isstringlike(value):
            raise TypeError("must assign a string")
        elif self.readonly:
            raise AttributeError("control '%s' is readonly" % self.name)
        elif self.disabled:
            raise AttributeError("control '%s' is disabled" % self.name)
        self._value = value
    def pairs(self): return []
    def _click(self, form, coord, return_type):
        # Relative URL for ISINDEX submission: instead of "foo=bar+baz",
        # want "bar+baz".
        # This doesn't seem to be specified in HTML 4.01 spec. (ISINDEX is
        # deprecated in 4.01, but it should still say how to submit it).
        # Submission of ISINDEX is explained in the HTML 3.2 spec, though.
        base = urljoin(self.node.parentNode._htmlform.base_uri, form.action)
        url = urljoin(base, "?"+urllib.quote_plus(self.value))
        req_data = url, None, []

        if return_type == "pairs":
            return []
        elif return_type == "request_data":
            return req_data
        else:
            return urllib2.Request(url)
    def __str__(self):
        value = self.value
        if value is None: value = "<None>"

        infos = []
        if self.disabled: infos.append("disabled")
        if self.readonly: infos.append("readonly")
        info = ", ".join(infos)
        if info: info = " (%s)" % info

        return "<%s(%s)%s>" % (self.__class__.__name__, value, info)


class SubmitControl(ScalarControl):
    """
    Covers:

    INPUT/SUBMIT
    BUTTON/SUBMIT

    """
    def __init__(self, node):
        ScalarControl.__init__(self, node)
        self.readonly = False
        if self.value is None: self.value = ""
        self.readonly = True

    def _click(self, form, coord, return_type):
        self._clicked = coord
        r = form._switch_click(return_type)
        self._clicked = False
        return r

    def pairs(self):
        if self.disabled or not self._clicked:
            return []
        return ScalarControl.pairs(self)


class ButtonControl(SubmitControl):
    """
    Covers:

    BUTTON/SUBMIT
    BUTTON/BUTTON
    BUTTON/RESET

    """
    def _get_type(self): return button_type(self.node)


class ImageControl(SubmitControl):
    """
    Covers:

    INPUT/IMAGE

    The value attribute of an ImageControl is always None.  Coordinates are
    specified using one of the HTMLForm.click* methods.

    """
    method_attrs = copy.copy(ScalarControl.method_attrs)
    method_attrs.remove("disabled")
    method_attrs.append("value")
    readonly_attrs = copy.copy(ScalarControl.readonly_attrs)
    readonly_attrs.append("value")
    def __init__(self, node):
        ScalarControl.__init__(self, node)
        self.disabled = False
        self.readonly = False
    def _get_value(self): return None
    def _get_disabled(self): return None
    def pairs(self):
        clicked = self._clicked
        if self.disabled or not clicked:
            return []
        name = self.name
        if name is None: return []
        return [("%s.x" % name, clicked[0]),
                ("%s.y" % name, clicked[1])]


class ListControl(ReadonlyMixin, Control):
    """Control representing a sequence of items.

    The value attribute of a ListControl represents the selected list items in
    the control.

    ListControl implements both list controls that take a single value and
    those that take multiple values.

    ListControls accept sequence values only.  Some controls only accept
    sequences of length 0 or 1 (RADIO, and single-selection SELECT).
    In those cases, ItemCountError is raised if len(sequence) > 1.  CHECKBOXes
    and multiple-selection SELECTs (those having the "multiple" HTML attribute)
    accept sequences of any length.

    Note the following mistake:

    control.value = some_value
    assert control.value == some_value    # not necessarily true

    The reason for this is that the value attribute always gives the list items
    in the order they were listed in the HTML.

    ListControl items can also be referred to by their labels instead of names.
    Use the by_label argument, and the set_value_by_label, get_value_by_label
    methods.

    XXX RadioControl and CheckboxControl don't implement by_label yet.

    Note that, rather confusingly, though SELECT controls are represented in
    HTML by SELECT elements (which contain OPTION elements, representing
    individual list items), CHECKBOXes and RADIOs are not represented by *any*
    element.  Instead, those controls are represented by a collection of INPUT
    elements.  For example, this is a SELECT control, named "control1":

    <select name="control1">
     <option>foo</option>
     <option value="1">bar</option>
    </select>

    and this is a CHECKBOX control, named "control2":

    <input type="checkbox" name="control2" value="foo" id="cbe1">
    <input type="checkbox" name="control2" value="bar" id="cbe2">

    The id attribute of a CHECKBOX or RADIO ListControl is always that of its
    first element (for example, "cbe1" above).


    Additional read-only public attribute: multiple.

    """
    readonly_attrs = ["type", "name", "id", "attrs"]
    method_attrs = ["type", "name", "id", "value", "disabled", "attrs"]
    def __init__(self, node, select_default=False,
                 called_as_base_class=False):
        if not called_as_base_class:
            raise NotImplementedError()

        self.node = node
        self.readonly = False

    def _find_item(self, name, by_label):
        items = self._items
        if by_label:
            items = [i for i in items if self._item_label(i) == name]
        else:
            items = [i for i in items if self._item_value(i) == name]
        if items:
            return items[0]
        else:
            if by_label: fmt = "no item labelled '%s'"
            else: fmt = "no item named '%s'"
            raise ItemNotFoundError(fmt % name)

    def _set_selected_state(self, name, action, by_label):
        from _js import fireHTMLEvent

        if not isstringlike(name):
            raise TypeError("item name must be string-like")
        if self.disabled:
            raise AttributeError("control '%s' is disabled" % self.name)
        if self.readonly:
            raise AttributeError("control '%s' is readonly" % self.name)

        item = self._find_item(name, by_label)
        if action == 2:
            action = not getattr(item, self.__class__.selected_attr)
        if action and self._item_disabled(item):
            raise AttributeError("item '%s' is disabled" % name)

        changed = False
        sa = self.__class__.selected_attr
        before = getattr(item, sa)
        if before != action:
            changed = True
        if not self.multiple:
            for i in self._items:
                setattr(i, sa, False)
        setattr(item, sa, action)
        if changed:
            fireHTMLEvent(self.node, "change")

    def _set_single_selected_state(self, action, by_label):
        items = self._items
        if len(items) != 1:
            raise ItemCountError("'%s' is not a single-item control" %
                                 self.name)

        item = items[0]
        if by_label: name = self._item_label(item)
        else: name = self._item_value(item)
        self._set_selected_state(name, action, by_label)

    def get_item_attrs(self, name, by_label=False):
        option = self._find_item(name, by_label)
        return attrs_dict_from_node(option)

    def _get_id(self):
        if self.node.hasAttribute("id"):
            return self.node.getAttribute("id")
        else:
            return None

    def _get_value(self):
        return [self._item_value(i) for i in self._items if
                getattr(i, self.__class__.selected_attr)]

    def _set_value(self, value):
        self._set_value_impl(value, False)
    def _set_value_impl(self, value, by_label):
        from _js import fireHTMLEvent

        if value is None or isstringlike(value):
            raise TypeError("ListControl, must set a sequence")

        if not self.multiple:
            if not (0 <= len(value) <= 1):
                raise ItemCountError("single selection list, must set "
                                     "sequence of length 0 or 1")

        if self.disabled:
            raise AttributeError("control '%s' is disabled" % self.name)
        if self.readonly:
            raise AttributeError("control '%s' is readonly" % self.name)

        items = self._items
        selected = [False]*len(items)

        for v in value:
            found = False
            for i in range(len(items)):
                item = items[i]
                if ((by_label and v == self._item_label(item)) or
                    (not by_label and v == self._item_value(item))):
                    if self._item_disabled(item):
                        raise AttributeError("item '%s' is disabled" % v)
                    selected[i] = True
                    found = True
                    break
            if not found:
                raise ItemNotFoundError("no item named '%s'" % repr(v))

        changed = False
        for i in range(len(items)):
            item = items[i]
            before = getattr(item, self.__class__.selected_attr)
            after = selected[i]
            if before != after:
                changed = True
                setattr(item, self.__class__.selected_attr, after)
        if changed:
            fireHTMLEvent(self.node, "change")

    def _item_value(self, item):
        return item.value
    def _item_label(self, item):
        return item.label

    def toggle(self, name, by_label=False):
        return self._set_selected_state(name, 2, by_label)
    def set(self, selected, name, by_label=False):
        action = int(bool(selected))
        return self._set_selected_state(name, action, by_label)

    def toggle_single(self, by_label=False):
        self._set_single_selected_state(2, by_label)
    def set_single(self, selected, by_label=False):
        action = int(bool(selected))
        self._set_single_selected_state(action, by_label)

    def _get_name(self):
        if self.node.hasAttribute("name"):
            return self.node.getAttribute("name")
        else:
            return None

    def get_value_by_label(self):
        raise NotImplementedError("control '%s' does not yet support "
                                  "by_label" % self.name)
    def set_value_by_label(self, value):
        raise NotImplementedError("control '%s' does not yet support "
                                  "by_label" % self.name)

    def possible_items(self, by_label=False):
        if by_label:
            #print "possible_items, label"
            return [self._item_label(i) for i in self._items]
        else:
            return [self._item_value(i) for i in self._items]

    def _get_attrs(self): return attrs_dict_from_node(self.node)

    def _item_str(self, i):
        item = self._items[i]
        item_name = self._item_value(item)
        if getattr(item, self.__class__.selected_attr):
            item_name = "*"+item_name
        if self._item_disabled(item):
            item_name = "(%s)" % item_name
        return item_name

    def pairs(self):
        if self.disabled:
            return []

        control_name = self.name  # usually the name HTML attribute
        pairs = []
        for item in self._items:
            if getattr(item, self.__class__.selected_attr):
                pairs.append((control_name, self._item_value(item)))
        return pairs

    def __str__(self):
        name = self.name
        if name is None: name = "<None>"

        display = []
        for i in range(len(self._items)):
            display.append(self._item_str(i))

        infos = []
        if self.disabled: infos.append("disabled")
        if self.readonly: infos.append("readonly")
        info = ", ".join(infos)
        if info: info = " (%s)" % info

        return "<%s(%s=[%s])%s>" % (self.__class__.__name__,
                                    name, ", ".join(display), info)


class SelectControl(ListControl):
    """
    Covers:

    SELECT (and OPTION)

    SELECT control values and labels are subject to some messy defaulting
    rules.  For example, if the HTML repreentation of the control is:

    <SELECT name=year>
      <OPTION value=0 label="2002">current year</OPTION>
      <OPTION value=1>2001</OPTION>
      <OPTION>2000</OPTION>
    </SELECT>

    The items, in order, have labels "2002", "2001" and "2000", whereas their
    values are "0", "1" and "2000" respectively.  Note that the value of the
    last OPTION in this example defaults to its contents, as specified by RFC
    1866, as do the labels of the second and third OPTIONs.

    The OPTION labels are sometimes more meaningful than the OPTION values,
    which can make for more maintainable code.

    Additional read-only public attribute: attrs

    The attrs attribute is a dictionary of the original HTML attributes of the
    SELECT element.  Other ListControls do not have this attribute, because in
    other cases the control as a whole does not correspond to any single HTML
    element.  The get_item_attrs method may be used as usual to get at the
    HTML attributes of the HTML elements corresponding to individual list items
    (for SELECT controls, these are OPTION elements).

    XXX I don't think I implemented this next bit (I should fix it, but you can
    use the DOM instead easily enough):

    Another special case is that the attributes dictionaries returned by
    get_item_attrs have a special key "contents" which does not correspond to
    any real HTML attribute, but rather contains the contents of the OPTION
    element:

    <OPTION>this bit</OPTION>

    """
    selected_attr = "selected"
    method_attrs = copy.copy(ListControl.method_attrs)
    method_attrs.append("multiple")

    def __init__(self, node, select_default=False):
        ListControl.__init__(self, node, select_default, True)
        self.multiple = node.multiple

        self._items = node.options

        if not self.multiple:
            # make sure only one item is selected
            found = False
            for i in self._items:
                if found:
                    i.selected = False
                    continue
                if i.selected: found = True

    def _get_multiple(self):
        return self.node.multiple
    def _set_multiple(self, value):
        self.node.multiple = value

    def _get_type(self): return "select"
    def _get_disabled(self):
        return self.node.disabled
    def _set_disabled(self, value):
        self.node.multiple = value

    # Unfortunately, since OPTGROUP is not explicitly part of
    # ClientForm model, but does view OPTIONs as inheriting their
    # disabled-ness from its OPTGROUP (if any), this is slightly odd.
    # This only messes with an OPTGROUP if it has to.
    def _item_disabled(self, item):
        parent = item.parentNode
        if parent.nodeName == "OPTGROUP":
            return (item.disabled or parent.disabled)
        else:
            return item.disabled
    def get_item_disabled(self, name, by_label=False):
        item = self._find_item(name, by_label)
        return self._item_disabled(item)
    def set_item_disabled(self, disabled, name, by_label=False):
        item = self._find_item(name, by_label)
        item.disabled = disabled

        if not disabled:
            parent = item.parentNode
            if parent.nodeName == "OPTGROUP":
                parent.disabled = disabled
    def set_all_items_disabled(self, disabled):
        for item in self._items:
            item.disabled = disabled

            if not disabled:
                parent = item.parentNode
                if parent.nodeName == "OPTGROUP":
                    parent.disabled = disabled

    def _item_value(self, item):
        if item.hasAttribute("value"):
            return item.value
        else:
            text = item.text
            if text is None:
                return None
            else:
                return text.strip()
    def _item_label(self, item):
        if item.hasAttribute("label"):
            return item.label
        else:
            text = item.text
            if text is None:
                return None
            else:
                return text.strip()

    def get_value_by_label(self):
        return [self._item_label(i) for i in self._items if
                getattr(i, self.__class__.selected_attr)]

    def set_value_by_label(self, value):
        if isstringlike(value):
            raise TypeError("ListControl, must set a sequence, not a string")
        if self.disabled:
            raise AttributeError("control '%s' is disabled" % self.name)
        if self.readonly:
            raise AttributeError("control '%s' is readonly" % self.name)

        self._set_value_impl(value, True)


class InputListControl(ListControl):
    method_attrs = copy.copy(ListControl.method_attrs)
    for attr in "disabled", "attrs":
        method_attrs.remove(attr)
    readonly_attrs = copy.copy(ListControl.readonly_attrs)
    readonly_attrs.remove("attrs")

    selected_attr = "checked"

    def __init__(self, node, select_default=False):
        ListControl.__init__(self, node, select_default, True)
        self._items = node.childNodes
        # default for value HTML attribute
        for i in self._items:
            if not i.hasAttribute("VALUE"):
                i.setAttribute("VALUE", "on")
        # use first input as representative of control
        self.node = self._items[0]
        self.multiple = (self.node.type == "Checkbox")
        self.disabled = False

    def _find_item(self, name, by_label):
        if by_label:
            raise NotImplementedError("control '%s' does not yet support "
                                      "by_label" % self.name)
        return ListControl._find_item(self, name, by_label)

    def _set_value_impl(self, value, by_label):
        if by_label:
            raise NotImplementedError("control '%s' does not yet support "
                                      "by_label" % self.name)
        return ListControl._set_value_impl(self, value, by_label)

    def _item_disabled(self, item):
        return item.disabled
    def get_item_disabled(self, name, by_label=False):
        item = self._find_item(name, by_label)
        return item.disabled
    def set_item_disabled(self, disabled, name, by_label=False):
        item = self._find_item(name, by_label)
        item.disabled = disabled
    def set_all_items_disabled(self, disabled):
        for item in self.node.items:
            item.disabled = disabled

    def possible_items(self, by_label=False):
        if by_label:
            raise NotImplementedError("control '%s' does not yet support "
                                      "by_label" % self.name)
        return ListControl.possible_items(self, by_label)

    def _set_single_selected_state(self, action, by_label):
        if by_label:
            raise NotImplementedError("control '%s' does not yet support "
                                      "by_label" % self.name)
        return ListControl._set_single_selected_state(self, action, by_label)


class CheckboxControl(InputListControl):
    """
    Covers:

    INPUT/CHECKBOX

    """
    def _get_type(self): return "checkbox"


class RadioControl(InputListControl):
    """
    Covers:

    INPUT/RADIO

    """
    def _get_type(self): return "radio"


def is_listcontrol(control): return hasattr(control, "possible_items")


class HTMLForm(ReadonlyMixin):
    """Represents a single HTML <form> ... </form> element.

    A form consists of a sequence of controls that usually have names, and
    which can take on various values.  The values of the various types of
    controls represent variously: text, zero-, one- or many-of-many choices,
    and files to be uploaded.

    Forms can be filled in with data to be returned to the server, and then
    submitted, using the click method to generate a request object suitable for
    passing to urllib2.urlopen (or the click_request_data or click_pairs
    methods if you're not using urllib2).

    import ClientForm
    forms = ClientForm.ParseFile(html, base_uri)
    form = forms[0]

    form["query"] = "Python"
    form.set("lots", "nr_results")

    response = urllib2.urlopen(form.click())

    Usually, HTMLForm instances are not created directly.  Instead, the
    ParseFile or ParseResponse factory functions are used.

    Indexing a form (form["control_name"]) returns the named Control's value
    attribute.  Assignment to a form index (form["control_name"] = something)
    is equivalent to assignment to the named Control's value attribute.  If you
    need to be more specific than just supplying the control's name, use the
    set_value and get_value methods.

    ListControl values are lists of item names.  The list item's name is the
    value of the corresponding HTML element's "value" attribute.

    Example:

      <INPUT type="CHECKBOX" name="cheeses" value="leicester"></INPUT>
      <INPUT type="CHECKBOX" name="cheeses" value="cheddar"></INPUT>

    defines a CHECKBOX control with name "cheeses" which has two items, named
    "leicester" and "cheddar".

    Another example:

      <SELECT name="more_cheeses">
        <OPTION>1</OPTION>
        <OPTION value="2" label="CHEDDAR">cheddar</OPTION>
      </SELECT>

    defines a SELECT control with name "more_cheeses" which has two items,
    named "1" and "2".

    To set, clear or toggle individual list items, use the set and toggle
    methods.  To set the whole value, do as for any other control:use indexing
    or the set_/get_value methods.

    Example:

    # select *only* the item named "cheddar"
    form["cheeses"] = ["cheddar"]
    # select "cheddar", leave other items unaffected
    form.set("cheddar", "cheeses")

    Some controls (RADIO and SELECT without the multiple attribute) can only
    have zero or one items selected at a time.  Some controls (CHECKBOX and
    SELECT with the multiple attribute) can have multiple items selected at a
    time.  To set the whole value of a multiple-selection ListControl, assign a
    sequence to a form index:

    form["cheeses"] = ["cheddar", "leicester"]

    To check whether a control has an item, or whether an item is selected,
    respectively:

    "cheddar" in form.possible_items("cheeses")
    "cheddar" in form["cheeses"]  # (or "cheddar" in form.get_value("cheeses"))

    Note that some items may be disabled (see below).

    Note the following mistake:

    form[control_name] = control_value
    assert form[control_name] == control_value  # not necessarily true

    The reason for this is that form[control_name] always gives the list items
    in the order they were listed in the HTML.

    List items (hence list values, too) can be referred to in terms of list
    item labels rather than list item names.  Currently, this is only possible
    for SELECT controls (this is a bug).  To use this feature, use the by_label
    arguments to the various HTMLForm methods.  Note that it is *item* names
    (hence ListControl values also), not *control* names, that can be referred
    to by label.

    The question of default values of OPTION contents, labels and values is
    somewhat complicated: see SelectControl.__doc__ and
    ListControl.get_item_attrs.__doc__ if you think you need to know.

    Controls can be disabled or readonly.  In either case, the control's value
    cannot be changed until you clear those flags (using the methods on
    HTMLForm).  Disabled is the state typically represented by browsers by
    `greying out' a control.  Disabled controls are not `successful' -- they
    don't cause data to get returned to the server.  Readonly controls usually
    appear in browsers as read-only text boxes.  Readonly controls are
    successful.  List items can also be disabled.  Attempts to select disabled
    items (with form[name] = value, or using the ListControl.set method, for
    example) fail.  Attempts to clear disabled items are allowed.

    If a lot of controls are readonly, it can be useful to do this:

    form.set_all_readonly(False)

    When you want to do several things with a single control, or want to do
    less common things, like changing which controls and items are disabled,
    you can get at a particular control:

    control = form.find_control("cheeses")
    control.set_item_disabled(False, "gruyere")
    control.set("gruyere")

    Most methods on HTMLForm just delegate to the contained controls, so see
    the docstrings of the various Control classes for further documentation.
    Most of these delegating methods take name, type, kind, id and nr arguments
    to specify the control to be operated on: see
    HTMLForm.find_control.__doc__.

    ControlNotFoundError (subclass of ValueError) is raised if the specified
    control can't be found.  This includes occasions where a non-ListControl
    is found, but the method (set, for example) requires a ListControl.
    ItemNotFoundError (subclass of ValueError) is raised if a list item can't
    be found.  ItemCountError (subclass of ValueError) is raised if an attempt
    is made to select more than one item and the control doesn't allow that, or
    set/get_single are called and the control contains more than one item.
    AttributeError is raised if a control or item is readonly or disabled and
    an attempt is made to alter its value.

    XXX CheckBoxControl and RadioControl don't yet support item access by label

    Security note: Remember that any passwords you store in HTMLForm instances
    will be saved to disk in the clear if you pickle them (directly or
    indirectly).  The simplest solution to this is to avoid pickling HTMLForm
    objects.


    Public attributes:

    action: full (absolute URI) form action
    method: "GET" or "POST"
    enctype: form transfer encoding MIME type
    name: name of form (None if no name was specified)
    attrs: dictionary mapping original HTML form attributes to their values

    controls: list of Control instances; do not alter this list
     (instead, call form.new_control to make a Control and add it to the
     form, or control.add_to_form if you already have a Control instance)



    Methods for form filling:
    -------------------------

    Most of the these methods have very similar arguments.  See
    HTMLForm.find_control.__doc__ for details of the name, type, kind and nr
    arguments.  See above for a description of by_label.

    def find_control(self,
                     name=None, type=None, kind=None, id=None, predicate=None,
                     nr=None)

    get_value(name=None, type=None, kind=None, id=None, nr=None,
              by_label=False)
    set_value(value,
              name=None, type=None, kind=None, id=None, nr=None,
              by_label=False)

    set_all_readonly(readonly)


    Methods applying only to ListControls:

    possible_items(name=None, type=None, kind=None, id=None, nr=None,
                   by_label=False)

    set(selected, item_name,
        name=None, type=None, kind=None, id=None, nr=None,
        by_label=False)
    toggle(item_name,
           name=None, type=None, id=None, nr=None,
           by_label=False)

    set_single(selected,
               name=None, type=None, kind=None, id=None, nr=None,
               by_label=False)
    toggle_single(name=None, type=None, kind=None, id=None, nr=None,
                  by_label=False)


    Method applying only to FileControls:

    add_file(file_object,
             content_type="application/octet-stream", filename=None,
             name=None, id=None, nr=None)


    Methods applying only to clickable controls:

    click(name=None, type=None, id=None, nr=0, coord=(1,1))
    click_request_data(name=None, type=None, id=None, nr=0, coord=(1,1))
    click_pairs(name=None, type=None, id=None, nr=0, coord=(1,1))

    """
    type2class = {
        "text": TextControl,
        # XXXX I'm sure there's something I need to do to PasswordControl,
        #   damned if I can remember what it is though...
        "password": PasswordControl,
        "hidden": HiddenControl,
        "textarea": TextareaControl,

        "isindex": IsindexControl,

        "file": FileControl,

        "submit": SubmitControl,
        # XXX these five may have issues to think through, since some were
        # ignored in ClientForm
        "button": ButtonControl,  # XXX
        "reset": SubmitControl,  # XXX
        "submitbutton": ButtonControl,  # XXX
        "buttonbutton": ButtonControl,  # XXX
        "resetbutton": ButtonControl,  # XXX
        "image": ImageControl,

        "radio": RadioControl,
        "checkbox": CheckboxControl,
        "select": SelectControl,
        }

    readonly_attrs = ["attrs"]
    method_attrs = ["action", "method", "enctype", "name", "attrs"]

#---------------------------------------------------
# Initialisation.  Use ParseResponse / ParseFile instead.

    def __init__(self, base_uri, node):
        if not isstringlike(base_uri):
            raise TypeError("base_uri must be string-like")
        self.base_uri = base_uri
        self.node = node
        if node.method == "": node.method = "GET"
        if node.encType == "": node.encType = "application/x-www-form-urlencoded"

        self.controls = []
        frag = node.ownerDocument.createDocumentFragment()
        inputlistcontrol = None

        for cn in control_node_generator(node):
            if (cn.nodeName == "INPUT" and
                cn.type in ["Radio", "Checkbox"]):
                id = (cn.name, cn.type)
                if inputlistcontrol != id:
                    # start building a new INPUT listcontrol
                    if inputlistcontrol is not None:
                        self._add_control(frag)  # end the current one first
                        frag = node.ownerDocument.createDocumentFragment()
                    inputlistcontrol = id
                nn = cn.cloneNode(False)
                frag.appendChild(nn)
                continue
            elif inputlistcontrol:
                # start of new control: ends the current INPUT listcontrol
                self._add_control(frag)
                frag = node.ownerDocument.createDocumentFragment()
                inputlistcontrol = None
            self._add_control(cn)
        else:
            if inputlistcontrol:
                # </form> inside an INPUT list control
                self._add_control(frag)

    def _add_control(self, node):
        # dispatch control node to appropriate Control class
        name = node.nodeName.lower()
        assert name != ""
        if name in ["select", "textarea", "isindex"]:
            ctype = name
        elif name == "button":
            ctype = button_type(node)
        elif name == "input":
            ctype = node.type.lower()
            if ctype == "":
                ctype = "text"
                # have to actually set TYPE HTML attribute to the
                # default value (text), or we won't be able to set
                # readOnly later
                node.setAttribute("TYPE", "Text")
        elif name == "#document-fragment":
            ctype = node.childNodes[0].type.lower()
            assert ctype in ["radio", "checkbox"]
        else:
            assert False, name

        klass = self.__class__.type2class[ctype]
        control = klass(node)
        node._control = control
        self.controls.append(control)

#---------------------------------------------------
    def __str__(self):
        action = urljoin(self.base_uri, self.action)
        header = "%s %s %s" % (self.method, action, self.enctype)
        rep = [header]
        for control in self.controls:
            rep.append("  %s" % str(control))
        return "<%s>" % "\n".join(rep)

#---------------------------------------------------
# Attribute access

    def _get_action(self): return self.node.action
    def _set_action(self, value): self.node.action = value
    def _get_method(self): return self.node.method.upper()
    def _set_method(self, value): self.node.method = value
    def _get_enctype(self): return self.node.encType.lower()
    def _set_enctype(self, value): self.node.encType = value
    def _get_name(self):
        if self.node.hasAttribute("name"):
            return self.node.getAttribute("name")
        else:
            return None
    def _set_name(self, value): self.node.name = value
    def _get_attrs(self): return attrs_dict_from_node(self.node)

#---------------------------------------------------
# Form-filling methods.

    def __getitem__(self, name):
        return self.find_control(name).value
    def __setitem__(self, name, value):
        control = self.find_control(name)
        try:
            control.value = value
        except AttributeError, e:
            raise ValueError(str(e))

    def get_value(self,
                  name=None, type=None, kind=None, id=None, nr=None,
                  by_label=False):
        """Return value of control.

        If type and nr are not supplied, equivalent to form[name].

        """
        c = self.find_control(name, type, kind, id, nr=nr)
        if by_label:
            try:
                meth = c.get_value_by_label
            except AttributeError:
                raise NotImplementedError(
                    "control '%s' does not yet support by_label" % c.name)
            else:
                return meth()
        else:
            return c.value
    def set_value(self, value,
                  name=None, type=None, kind=None, id=None, nr=None,
                  by_label=False):
        """Set value of control.

        If type and nr are not supplied, equivalent to form[name] = value.

        """
        c = self.find_control(name, type, kind, id, nr=nr)
        if by_label:
            try:
                meth = c.set_value_by_label
            except AttributeError:
                raise NotImplementedError(
                    "control '%s' does not yet support by_label" % c.name)
            else:
                meth(value)
        else:
            c.value = value

    def set_all_readonly(self, readonly):
        for control in self.controls:
            control.readonly = bool(readonly)


#---------------------------------------------------
# Form-filling methods applying only to ListControls.

    def possible_items(self,
                       name=None, type=None, kind=None, id=None, nr=None,
                       by_label=False):
        """Return a list of all values that the specified control can take."""
        c = self._find_list_control(name, type, kind, id, nr)
        return c.possible_items(by_label)

    def set(self, selected, item_name,
            name=None, type=None, kind=None, id=None, nr=None,
            by_label=False):
        """Select / deselect named list item.

        selected: boolean selected state

        """
        self._find_list_control(name, type, kind, id, nr).set(
            selected, item_name, by_label)
    def toggle(self, item_name,
               name=None, type=None, kind=None, id=None, nr=None,
               by_label=False):
        """Toggle selected state of named list item."""
        self._find_list_control(name, type, kind, id, nr).toggle(
            item_name, by_label)

    def set_single(self, selected,
                   name=None, type=None, kind=None, id=None, nr=None,
                   by_label=False):
        """Select / deselect list item in a control having only one item.

        If the control has multiple list items, ItemCountError is raised.

        This is just a convenience method, so you don't need to know the item's
        name -- the item name in these single-item controls is usually
        something meaningless like "1" or "on".

        For example, if a checkbox has a single item named "on", the following
        two calls are equivalent:

        control.toggle("on")
        control.toggle_single()

        """
        self._find_list_control(name, type, kind, id, nr).set_single(
            selected, by_label)
    def toggle_single(self, name=None, type=None, kind=None, id=None, nr=None,
                      by_label=False):
        """Toggle selected state of list item in control having only one item.

        The rest is as for HTMLForm.set_single.__doc__.

        """
        self._find_list_control(name, type, kind, id, nr).toggle_single(by_label)

#---------------------------------------------------
# Form-filling method applying only to FileControls.

    def add_file(self, file_object, content_type=None, filename=None,
                 name=None, id=None, nr=None):
        """Add a file to be uploaded.

        file_object: file-like object (with read method) from which to read
         data to upload
        content_type: MIME content type of data to upload
        filename: filename to pass to server

        If filename is None, no filename is sent to the server.

        If content_type is None, the content type is guessed based on the
        filename and the data from read from the file object.

        XXX
        At the moment, guessed content type is always application/octet-stream.
        Use sndhdr, imghdr modules.  Should also try to guess HTML, XML, and
        plain text.

        """
        self.find_control(name, "file", id=id, nr=nr).add_file(
            file_object, content_type, filename)

#---------------------------------------------------
# Form submission methods, applying only to clickable controls.

    def click(self, name=None, type=None, id=None, nr=0, coord=(1,1)):
        """Return request that would result from clicking on a control.

        The request object is a urllib2.Request instance, which you can pass to
        urllib2.urlopen (or ClientCookie.urlopen).

        Only some control types (INPUT/SUBMIT & BUTTON/SUBMIT buttons and
        IMAGEs) can be clicked.

        Will click on the first clickable control, subject to the name, type
        and nr arguments (as for find_control).  If no name, type, id or number
        is specified and there are no clickable controls, a request will be
        returned for the form in its current, un-clicked, state.

        IndexError is raised if a name, type and/or nr is specified but no
        matching control is found.  ValueError is raised if the HTMLForm has an
        enctype attribute that is not recognised.

        You can optionally specify a coordinate to click at, which only makes a
        difference if you clicked on an image.

        """
        from _js import fireHTMLEvent
        fireHTMLEvent(self.node, "submit")
        return self._click(name, type, id, nr, coord, "request")

    def click_request_data(self,
                           name=None, type=None, id=None, nr=0, coord=(1,1)):
        """As for click method, but return a tuple (url, data, headers).

        You can use this data to send a request to the server.  This is useful
        if you're using httplib or urllib rather than urllib2.  Otherwise, use
        the click method.

        # Untested.  Have to subclass to add headers, I think -- so use urllib2
        # instead!
        import urllib
        url, data, hdrs = form.click_request_data()
        r = urllib.urlopen(url, data)

        # Untested.  I don't know of any reason to use httplib -- you can get
        # just as much control with urllib2.
        import httplib, urlparse
        url, data, hdrs = form.click_request_data()
        tup = urlparse(url)
        host, path = tup[1], urlparse.urlunparse((None, None)+tup[2:])
        conn = httplib.HTTPConnection(host)
        if data:
            httplib.request("POST", path, data, hdrs)
        else:
            httplib.request("GET", path, headers=hdrs)
        r = conn.getresponse()

        """
        return self._click(name, type, id, nr, coord, "request_data")

    def click_pairs(self, name=None, type=None, id=None, nr=0, coord=(1,1)):
        """As for click_request_data, but returns a list of (key, value) pairs.

        You can use this list as an argument to ClientForm.urlencode.  This is
        usually only useful if you're using httplib or urllib rather than
        urllib2 or ClientCookie.  It may also be useful if you want to manually
        tweak the keys and/or values, but this should not be necessary.
        Otherwise, use the click method.

        Note that this method is only useful for forms of MIME type
        x-www-form-urlencoded.  In particular, it does not return the
        information required for file upload.  If you need file upload and are
        not using urllib2, use click_request_data.

        Also note that Python 2.0's urllib.urlencode is slightly broken: it
        only accepts a mapping, not a sequence of pairs, as an argument.  This
        messes up any ordering in the argument.  Use ClientForm.urlencode
        instead.

        """
        return self._click(name, type, id, nr, coord, "pairs")

#---------------------------------------------------

    def find_control(self,
                     name=None, type=None, kind=None, id=None, predicate=None,
                     nr=None):
        """Locate some specific control within the form.

        At least one of the name, type, kind, predicate and nr arguments must
        be supplied.  If no matching control is found, ControlNotFoundError is
        raised.

        If name is specified, then the control must have the indicated name.

        If type is specified then the control must have the specified type (in
        addition to the types possible for <input> HTML tags: "text",
        "password", "hidden", "submit", "image", "button", "radio", "checkbox",
        "file" we also have "reset", "buttonbutton", "submitbutton",
        "resetbutton", "textarea", "select" and "isindex").

        If kind is specified, then the control must fall into the specified
        group, each of which satisfies a particular interface.  The types are
        "text", "list", "multilist", "singlelist", "clickable" and "file".

        If id is specified, then the control must have the indicated id.

        If predicate is specified, then the control must match that function.
        The predicate function is passed the control as its single argument,
        and should return a boolean value indicating whether the control
        matched.

        nr, if supplied, is the sequence number of the control (where 0 is the
        first).  Note that control 0 is the first control matching all the
        other arguments (if supplied); it is not necessarily the first control
        in the form.

        """
        if ((name is None) and (type is None) and (kind is None) and
            (id is None) and (predicate is None) and (nr is None)):
            raise ValueError(
                "at least one argument must be supplied to specify control")
        if nr is None: nr = 0

        return self._find_control(name, type, kind, id, predicate, nr)

#---------------------------------------------------
# Private methods.

    def _find_list_control(self,
                           name=None, type=None, kind=None, id=None, nr=None):
        if ((name is None) and (type is None) and (kind is None) and
            (id is None) and (nr is None)):
            raise ValueError(
                "at least one argument must be supplied to specify control")
        if nr is None: nr = 0

        return self._find_control(name, type, kind, id, is_listcontrol, nr)

    def _find_control(self, name, type, kind, id, predicate, nr):
        if (name is not None) and not isstringlike(name):
            raise TypeError("control name must be string-like")
        if (type is not None) and not isstringlike(type):
            raise TypeError("control type must be string-like")
        if (kind is not None) and not isstringlike(kind):
            raise TypeError("control kind must be string-like")
        if (id is not None) and not isstringlike(id):
            raise TypeError("control id must be string-like")
        if (predicate is not None) and not callable(predicate):
            raise TypeError("control predicate must be callable")
        if nr < 0: raise ValueError("control number must be a positive "
                                    "integer")

        orig_nr = nr

        for control in self.controls:
            if name is not None and name != control.name:
                continue
            if type is not None and type != control.type:
                continue
            if (kind is not None and
                not self._is_control_in_kind(control, kind)):
                continue
            if id is not None and id != control.id:
                continue
            if predicate and not predicate(control):
                continue
            if nr:
                nr = nr - 1
                continue
            return control

        description = []
        if name is not None: description.append("name '%s'" % name)
        if type is not None: description.append("type '%s'" % type)
        if kind is not None: description.append("kind '%s'" % kind)
        if id is not None: description.append("id '%s'" % id)
        if predicate is not None:
            description.append("matching predicate %s" % predicate)
        if orig_nr: description.append("nr %d" % orig_nr)
        description = ", ".join(description)
        raise ControlNotFoundError("no control with "+description)

    def _is_control_in_kind(self, control, kind):
        # XXX not OO
        if kind  == "list":
            return isinstance(control, ListControl)
        elif kind == "multilist":
            return bool(isinstance(control, ListControl) and control.multiple)
        elif kind == "singlelist":
            return bool(isinstance(control, ListControl) and
                        not control.multiple)
        elif kind == "file":
            return isinstance(control, FileControl)
        elif kind == "text":
            return isinstance(control, TextControl)
        elif kind == "clickable":
            return (isinstance(control, SubmitControl) or
                    isinstance(control, IsindexControl) or
                    isinstance(control, ImageControl))
        else:
            raise ValueError("no such control kind '%s'" % kind)

    def _click(self, name, type, id, nr, coord, return_type):
        try:
            control = self._find_control(name, type, "clickable", id, None, nr)
        except ControlNotFoundError:
            if ((name is not None) or (type is not None) or (id is not None) or
                (nr != 0)):
                raise
            # no clickable controls, but no control was explicitly requested by
            # name or number, so return state without clicking any control
            return self._switch_click(return_type)
        else:
            return control._click(self, coord, return_type)

    def _pairs(self):
        """Return sequence of (key, value) pairs suitable for urlencoding."""
        pairs = []
        for control in self.controls:
            pairs.extend(control.pairs())
        return pairs

    def _request_data(self):
        """Return a tuple (url, data, headers)."""
        method = self.method.upper()
        action = urljoin(self.base_uri, self.action)
        if method == "GET":
            if self.enctype != "application/x-www-form-urlencoded":
                raise ValueError(
                    "unknown GET form encoding type '%s'" % self.enctype)
            uri = "%s?%s" % (action, urllib.urlencode(self._pairs()))
            return uri, None, []
        elif method == "POST":
            if self.enctype == "application/x-www-form-urlencoded":
                return (action, urllib.urlencode(self._pairs()),
                        [("Content-type", self.enctype)])
            elif self.enctype == "multipart/form-data":
                data = StringIO()
                http_hdrs = []
                mw = MimeWriter(data, http_hdrs)
                f = mw.startmultipartbody("form-data", add_to_http_hdrs=True)
                for control in self.controls:
                    control._write_mime_data(mw)
                mw.lastpart()
                return action, data.getvalue(), http_hdrs
            else:
                raise ValueError(
                    "unknown POST form encoding type '%s'" % self.enctype)
        else:
            raise ValueError("Unknown method '%s'" % method)

    def _switch_click(self, return_type):
        # This is called by HTMLForm and clickable Controls to hide switching
        # on return_type.
        # XXX
        # not OO
        # duplicated in IsindexControl._click
        if return_type == "pairs":
            return self._pairs()
        elif return_type == "request_data":
            return self._request_data()
        else:
            req_data = self._request_data()
            req = urllib2.Request(req_data[0], req_data[1])
            for key, val in req_data[2]:
                req.add_header(key, val)
            return req


def form_node_generator(node):
    if node.nodeName == "FORM":
        yield node
    for node in node.childNodes:
        for node in form_node_generator(node):
            yield node


def control_node_generator(node):
    if node.nodeName in ["INPUT", "SELECT", "TEXTAREA", "BUTTON", "ISINDEX"]:
        yield node
    for node in node.childNodes:
        for node in control_node_generator(node):
            yield node

def decorate_DOM(window):
    """Decorate DOM with attributes supporting ClientForm interface.

    Generates HTMLForms from a DOM Document object and updates attributes on
    DOM pointing to ClientForm objects.

    """
    document = window.document
    forms = []
    for fn in form_node_generator(document):
        form = HTMLForm(window.location, fn)
        forms.append(form)
        fn._htmlform = form
    window._set_property("_htmlforms", forms)
