########################################################################
#
# File Name:            NodeList.py
#
#
"""

4DOM implementation has been replaced with a modified version of code from
Andrew Clover's pxdom here, in order to implement 'live'-ness of
HTMLCollections.

WWW: http://4suite.com/4DOM         e-mail: support@4suite.com

Copyright (c) 2000 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.com/COPYRIGHT  for license and copyright information
"""

import UserList

from xml.dom import NoModificationAllowedErr, Node

class _NONS:
  """ Singleton class used internally as a value for namespaceURI signifying
      that a non-namespace version of a node or method is in use; the
      accompanying localName is then the entire name. This is different to
      None, which is a null namespace value.
  """
  pass
NONS= _NONS()


class DOMObject:
  """ Base class that allows access to properties through calling getter and
      setter methods of the form _get_propertyName. Immutable properties can
      be made by providing no _set_propertyName method; immutable objects are
      made by setting the readonly property.
  """
  def __init__(self, readonly= False):
    self._readonly= readonly

  def _get_readonly(self):
    return self._readonly
  def _set_readonly(self, value):
    self._readonly= value

  def __getattr__(self, key):
    if key[:1]=='_':
      raise AttributeError, key
    try:
      getter= getattr(self, '_get_'+key)
    except AttributeError:
      raise AttributeError, key
    return getter()

  def __setattr__(self, key, value):
    if key[:1]=='_':
      self.__dict__[key]= value
      return
    if self._readonly and key not in ('readonly', 'nodeValue'):
      raise NoModificationAllowedErr(key)
    try:
      setter= getattr(self, '_set_'+key)
    except AttributeError:
      if hasattr(self, '_get_'+key):
        raise NoModificationAllowedErr(key)
      raise AttributeError, key
    setter(value)


class DOMList(DOMObject):
  """ A list structure that can be accessed either using the DOM IDL methods
      or Python list accessor constructs.
  """
  def __init__(self):
    DOMObject.__init__(self)
    self._list= []

  # DOM-style methods
  #
  def _get_length(self):
    return len(self._list)

  def item(self, index):
    if index<0 or index>=len(self._list):
      return None
    return self._list[index]

  # Python-style methods
  #
  def __len__(self):
    return len(self._list)

  def __getitem__(self, index):
    return self._list[index]

  def __setitem__(self, index, value):
    raise NoModificationAllowedErr('item(%s)' % str(index))

  def __delitem__(self, index):
    raise NoModificationAllowedErr('item(%s)' % str(index))

  # Mutable sequence convenience methods for internal use
  #
  def _index(self, value):
    return self._list.index(value)

  def _append(self, value):
    if self._readonly:
      raise NoModificationAllowedErr('item(%s)' % str(len(self._list)))
    self._list.append(value)

  def _insertseq(self, index, values):
    if self._readonly:
      raise NoModificationAllowedErr('item(%s)' % str(index))
    self._list[index:index]= values


class NodeList(DOMList):
  """ Abstract list of nodes dependent on an owner node.
  """
  def __init__(self, ownerNode= None):
    DOMList.__init__(self)
    self._ownerNode= ownerNode


class NodeListByTagName(NodeList):
  """ A NodeList returned by an Element.getElementsByTagName[NS] method. This
      is still live - the internal _list acts only as a cache, and is
      recalculated if the owner Element's contents have changed since it was
      last built.
  """
  def __init__(self, ownerNode, namespaceURI, localName):
      NodeList.__init__(self, ownerNode)
      self._namespaceURI= namespaceURI
      self._localName= localName
      self._sequence= None

  # XXXX very doubtful about these implementations: eg. exception handling
  def __setitem__(self, index, value):
      if self._sequence!=self._ownerNode._sequence: self._calculate()
      parent = self._ownerNode
      if index >= len(self._list):
        parent.appendChild(value)
      else:
        node = self._list[index]
        parent.replaceChild(value, node)
      parent._changed()
##       node = self._list[index]
##       parent = node.parentNode
##       parent.replaceChild(value, node)
##       parent._changed()

  def __delitem__(self, index):
      if self._sequence!=self._ownerNode._sequence: self._calculate()
      node = self._list[index]
      parent = node.parentNode
      parent.removeChild(node)
      parent._changed()

  def _get_length(self):
      if self._sequence!=self._ownerNode._sequence: self._calculate()
      return NodeList._get_length(self)

  def item(self, index):
      if self._sequence!=self._ownerNode._sequence: self._calculate()
      return NodeList.item(self, index)

  def __getitem__(self, index):
      if self._sequence!=self._ownerNode._sequence: self._calculate()
      return NodeList.__getitem__(self, index)

  def __len__(self):
      if self._sequence!=self._ownerNode._sequence: self._calculate()
      return NodeList.__len__(self)

  def _calculate(self):
      """ Recalculate the list. This method does the actual work of the
          Element.getElementsByTagName call.
      """
      self._list= []
      self._walk(self._ownerNode)
      self._sequence= self._ownerNode._sequence

  def _walk(self, element):
      """ Recursively add a node's child elements to the internal node list
          when they match the conditions passed to
          Element.getElementsByTagName.
      """
      for childNode in element.childNodes:
          if childNode.nodeType==Node.ELEMENT_NODE:
              if (
                  self._localName=='*' and
                  self._namespaceURI in ('*', NONS, childNode.namespaceURI)
                  ) or (
                  self._namespaceURI=='*' and
                  self._localName==childNode.localName
                  ) or (
                  self._namespaceURI==NONS and
                  self._localName==childNode.nodeName
                  ) or (
                  self._namespaceURI==childNode.namespaceURI and
                  self._localName==childNode.localName
                  ):
                  self._list.append(childNode)
              if (childNode.nodeType in
                  (Node.ELEMENT_NODE,Node.ENTITY_REFERENCE_NODE)):
                self._walk(childNode)


class HTMLOptionsCollection(NodeListByTagName):
  """Live list of OPTIONs, for internal use by ClientForm implementation.

  The need for this arises because HTMLSelectElement.options needs to contain
  the OPTIONs inside OPTGROUPs, not just those immediately under the SELECT.

  """
  def __init__(self, ownerNode):
    NodeList.__init__(self, ownerNode)
    self._sequence = None

  def namedItem(self, name):
    # XXXX XHTML should be case-sensitive, HTML not: is it?
    if self._sequence!=self._ownerNode._sequence: self._calculate()
    # check by id
    for cn in self._ownerNode.childNodes:
      if cn.nodeType == Node.ELEMENT_NODE:
        if cn.getAttribute(id) == name:
          return cn
    for cn in self._ownerNode.childNodes:
      if cn.nodeType == Node.ELEMENT_NODE:
        # XXXX only supposed to look at nodes that are allowed name attribute
        if cn.getAttribute("NAME") == name:
          return cn
    # XXXX presumably 'null' is None?

  def _set_length(self, value):
    if self._sequence!=self._ownerNode._sequence: self._calculate()
    int(value)
    #from xml.dom.ext import XHtmlPrettyPrint as pp
    #print "value", value
    #print "self._list", self._list
    for cn in self._list[value:]:
      #print "removing"
      #print "cn", cn
      #pp(cn)
      #print
      cn.parentNode.removeChild(cn)

  def _walk(self, element):
    for cn in element.childNodes:
      if cn.nodeType == Node.ELEMENT_NODE:
        if cn.nodeName == "OPTION":
          self._list.append(cn)
        elif cn.nodeName == "OPTGROUP":
          for opt in cn.childNodes:
            if opt.nodeType == Node.ELEMENT_NODE:
              assert opt.nodeName == "OPTION"
              self._list.append(opt)


## class NodeList(UserList.UserList):
##     def __init__(self, list=None):
##         UserList.UserList.__init__(self, list)
##         return

##     ### Attribute Access Methods ###

##     def __getattr__(self, name):
##         if name == 'length':
##             return len(self)
##         #Pass-through
##         return getattr(NodeList, name)

##     def __setattr__(self, name, value):
##         if name == 'length':
##             raise NoModificationAllowedErr()
##         #Pass-through
##         self.__dict__[name] = value

##     ### Attribute Methods ###

##     def _get_length(self):
##         return len(self)

##     ### Methods ###

##     def item(self, index):
##         if 0 <= index < len(self):
##             return self[int(index)]
##         return None

##     #Not defined in the standard
##     def contains(self, node):
##         return node in self

##     def __repr__(self):
##         st = "<NodeList at %x: [" % id(self)
##         if len(self):
##             for i in self[:-1]:
##                 st = st + repr(i) + ', '
##             st = st + repr(self[-1])
##         st = st + ']>'
##         return st
